@@ -48,25 +48,16 @@ class Document(Element):  # ToDo: add multi document support
 
     @classmethod
     def _read_name(cls, stream: nip.stream.Stream):
-        # possible document starts: "--- {name} {!tag}"
-        pos, operator = tokens.Operator.read(stream)
-        if pos > 0 and operator == '---':
-            stream.move(pos)
-        else:
-            return None
-
-        pos, indent = tokens.Indent.read(stream)
-        if indent is not None:  # new string means start of right value
-            return None
-
-        pos, doc_name = tokens.Name.read(stream)
-        if not pos:
-            return None
-        stream.move(pos)
-        return doc_name
+        read_tokens = stream.peek(tokens.Operator('---'), tokens.String) or \
+                      stream.peek(tokens.Operator('---'))
+        if read_tokens is not None:
+            stream.step()
+            if len(read_tokens) == 2:
+                return read_tokens[1].value
+        return ''
 
     def dump(self, dumper: nip.dumper.Dumper):
-        return "--- " + self.value.dump(dumper)
+        return "--- " + self.name + " " + self.value.dump(dumper)
 
 
 class RightValue(Element):
@@ -78,9 +69,9 @@ class RightValue(Element):
                 Iter.read(stream, parser) or \
                 Args.read(stream, parser) or \
                 FString.read(stream, parser) or \
-                Value.read(stream, parser) or \
+                Nothing.read(stream, parser) or \
                 InlinePython.read(stream, parser) or \
-                Nothing.read(stream, parser)
+                Value.read(stream, parser)
 
         if not value:
             raise nip.parser.ParserError(stream, "Wrong right value")
@@ -90,28 +81,26 @@ class RightValue(Element):
 
 class Value(Element):
     @classmethod
-    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Value:
-        return cls._read_token(tokens.Number, stream) or \
-               cls._read_token(tokens.Bool, stream) or \
-               cls._read_token(tokens.String, stream) or \
-               cls._read_token(tokens.List, stream) or \
-               cls._read_token(tokens.Dict, stream)
-
-    @staticmethod
-    def _read_token(cls: Type[tokens.Token], stream: nip.stream.Stream) -> Union[Value, None]:
-        pos, value = cls.read(stream)
-        if pos > 0:
-            stream.move(pos)
-            return Value(cls.__name__, value)
+    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[None, Value]:
+        tokens_list = [
+            tokens.Number,
+            tokens.Bool,
+            tokens.String,
+            tokens.List,
+            tokens.Dict
+        ]
+        for token in tokens_list:
+            read_tokens = stream.peek(token)
+            if read_tokens is not None:
+                stream.step()
+                return Value(read_tokens[0].name, read_tokens[0].value)
 
         return None
 
     def to_python(self):
         return self.value
 
-    def construct(self, constructor: nip.constructor.Constructor, always_pair=False):
-        if always_pair:
-            return [self.value], {}
+    def construct(self, constructor: nip.constructor.Constructor):
         return self.value
 
     def dump(self, dumper: nip.dumper.Dumper):
@@ -123,16 +112,14 @@ class Value(Element):
 class LinkCreation(Element):
     @classmethod
     def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Element, None]:
-        pos, op = tokens.Operator.read(stream)
-        if pos > 0 and op == '&':
-            stream.move(pos)
-        else:
+        read_tokens = stream.peek(tokens.Operator('&'), tokens.Name)
+        if read_tokens is None:
             return None
+        # if not isinstance(stream[1], tokens.Name):
+        #     raise nip.parser.ParserError(stream, "Wrong var creation")  # mb: No good errors now =(
 
-        pos, name = tokens.Name.read(stream)
-        if not pos:
-            raise nip.parser.ParserError(stream, "Wrong link creation")
-        stream.move(pos)
+        name = read_tokens[1].value
+        stream.step()
 
         value = RightValue.read(stream, parser)
         parser.links.append(name)
@@ -150,18 +137,14 @@ class LinkCreation(Element):
 class Link(Element):
     @classmethod
     def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Element, None]:
-        pos, op = tokens.Operator.read(stream)
-        if pos > 0 and (op == '*'):
-            stream.move(pos)
-        else:
+        read_tokens = stream.peek(tokens.Operator('*'), tokens.Name)
+        if read_tokens is None:
             return None
 
-        pos, name = tokens.Name.read(stream)
-        if not pos:
-            raise nip.parser.ParserError(stream, "Wrong link usage")
+        name = read_tokens[1].value
         if name not in parser.links:
             nip.parser.ParserError(stream, "Link usage before assignment")
-        stream.move(pos)
+        stream.step()
 
         return Link(name)
 
@@ -178,19 +161,13 @@ class Link(Element):
 class Tag(Element):
     @classmethod
     def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Tag, None]:
-        pos, op = tokens.Operator.read(stream)
-        if pos > 0 and op == '!':
-            stream.move(pos)
-        else:
+        read_tokens = stream.peek(tokens.Operator('!'), tokens.Name)
+        if read_tokens is None:
             return None
-
-        pos, name = tokens.Name.read(stream)
-        if not pos:
-            raise nip.parser.ParserError(stream, "Wrong tag creation")
-        stream.move(pos)
+        name = read_tokens[1].value
+        stream.step()
 
         value = RightValue.read(stream, parser)
-        #assert isinstance(value, (Args, Value)), "Tag should be created with List or Dict or Args"
 
         return Tag(name, value)
 
@@ -215,83 +192,53 @@ class Tag(Element):
 class Args(Element):
     @classmethod
     def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Args, None]:
-        pos, indent = tokens.Indent.read(stream)
-        if indent is None or indent <= stream.last_indent:
+        start_indent = stream.pos
+        if start_indent <= parser.last_indent:
             return None
 
-        args: list = cls._read_list(stream, parser, indent)
-        stream.pos = 0
-        kwargs: dict = cls._read_dict(stream, parser, indent)
+        args = []
+        kwargs = {}
+        while stream and stream.pos == start_indent:
+            parser.last_indent = start_indent
+
+            item = cls._read_list_item(stream, parser)
+            if item is not None:
+                args.append(item)
+                continue
+
+            key, value = cls._read_dict_pair(stream, parser)
+            if key is not None:
+                kwargs[key] = value
+                continue
+
+            break
 
         if not args and not kwargs:
-            raise nip.parser.ParserError(stream, "Error reading indented element")
+            return None
         return Args("args", (args, kwargs))
 
-    @classmethod
-    def _read_list(cls, stream: nip.stream.Stream, parser: nip.parser.Parser,
-                   start_indent: int) -> list:
-        pos, indent = tokens.Indent.read(stream)
-        current_list = []
-        while stream and indent == start_indent:
-            stream.move(pos)
-            stream.last_indent = indent
-            item = cls._read_list_item(stream, parser)
-            if not item:
-                break
-            current_list.append(item)
-
-            if not stream:
-                break
-
-            pos, indent = tokens.Indent.read(stream)
-
-        return current_list
 
     @classmethod
     def _read_list_item(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) \
             -> Union[Element, None]:
-        pos, op = tokens.Operator.read(stream)
-        if not pos or op != '-':
+        read_tokens = stream.peek(tokens.Operator('- '))
+        if read_tokens is None:
             return None
-        stream.move(pos)
+        stream.step()
 
         value = RightValue.read(stream, parser)
 
         return value
 
-    @classmethod
-    def _read_dict(cls, stream: nip.stream.Stream, parser: nip.parser.Parser,
-                   start_indent: int) -> dict:
-        pos, indent = tokens.Indent.read(stream)
-
-        current_dict = OrderedDict()
-        while stream and indent == start_indent:
-            stream.move(pos)
-            stream.last_indent = indent
-            key, value = cls._read_dict_pair(stream, parser)
-            if not key:
-                break
-            current_dict[key] = value
-
-            if not stream:
-                break
-
-            pos, indent = tokens.Indent.read(stream)
-
-        return current_dict
-
     @classmethod
     def _read_dict_pair(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) \
             -> Union[Tuple[str, Element], Tuple[None, None]]:
-        pos, key = tokens.Name.read(stream)
-        if not pos:
+        read_tokens = stream.peek(tokens.String, tokens.Operator(': '))
+        if read_tokens is None:
             return None, None
-        stream.move(pos)
 
-        pos, op = tokens.Operator.read(stream)
-        if not pos or op != ':':
-            raise nip.parser.ParserError(stream, "Wrong dict pair assertion")
-        stream.move(pos)
+        key = read_tokens[0].value
+        stream.step()
 
         value = RightValue.read(stream, parser)
 
@@ -334,13 +281,15 @@ class Args(Element):
             " "*dumper.indent + f"- {item.dump(dumper + dumper.default_shift)}"
             for item in self.value[0]
         ])
+        string = ('\n' if dumped_args else '') + dumped_args
 
         dumped_kwargs = '\n'.join([
             " "*dumper.indent + f"{key}: {value.dump(dumper + dumper.default_shift)}"
             for key, value in self.value[1].items()
         ])
+        string += ('\n' if dumped_kwargs else '') + dumped_kwargs
 
-        return '\n' + dumped_args + dumped_kwargs
+        return string
 
 
 class Iter(Element):
@@ -351,20 +300,16 @@ class Iter(Element):
 
     @classmethod
     def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Iter, None]:
-        pos, op = tokens.Operator.read(stream)
-        if pos > 0 and (op == '@'):
-            stream.move(pos)
-        else:
+        read_tokens = stream.peek(tokens.Operator('@'), tokens.Name, tokens.List) or \
+                      stream.peek(tokens.Operator('@'), tokens.List)
+        if read_tokens is None:
             return None
+        stream.step()
+        if len(read_tokens) == 2:
+            iterator = Iter('', read_tokens[1].value)
+        else:
+            iterator = Iter(read_tokens[1].value, read_tokens[2].value)
 
-        pos, name = tokens.Name.read(stream)
-        stream.move(pos)
-        # mb: read full RightValue here? what with nested cases?
-        pos, iter_list = tokens.List.read(stream)  # mb: or tuple
-        if not pos:
-            nip.parser.ParserError(stream, "Wrong iter creation")
-        stream.move(pos)
-        iterator = Iter(name, iter_list)
         parser.iterators.append(iterator)
         return iterator
 
@@ -375,7 +320,7 @@ class Iter(Element):
 
     def construct(self, constructor: nip.constructor.Constructor):
         if self.return_index == -1:
-            raise nip.constructor.ConstructorError("Iterator index was not specified by IterParser")
+            raise Exception("Iterator index was not specified by IterParser")
         return self.value[self.return_index]
 
     def dump(self, dumper: nip.dumper.Dumper):
@@ -388,13 +333,14 @@ class Iter(Element):
 
 class InlinePython(Element):
     @classmethod
-    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[InlinePython, None]:
-        pos, exec_string = tokens.InlinePython.read(stream)
-        if pos > 0:
-            stream.move(pos)
-            return InlinePython(value=exec_string)
-        else:
+    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> \
+            Union[InlinePython, None]:
+        read_tokens = stream.peek(tokens.InlinePython)
+        if read_tokens is None:
             return None
+        stream.step()
+        exec_string = read_tokens[0].value
+        return InlinePython(value=exec_string)
 
     def construct(self, constructor: nip.constructor.Constructor):
         locals().update(constructor.vars)
@@ -407,10 +353,15 @@ class InlinePython(Element):
 class Nothing(Element):
     @classmethod
     def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Nothing, None]:
-        pos, indent = tokens.Indent.read(stream)
-        if indent is None:
+        if not stream:
+            return Nothing()
+        if not stream.lines[stream.n][:stream.pos].isspace():
             return None
-        if indent <= stream.last_indent:
+        indent = stream.pos
+        # pos, indent = tokens.Indent.read(stream)
+        # if indent is None:  # mb this should be specifically processed
+        #     return None
+        if indent <= parser.last_indent:
             return Nothing()
 
     def construct(self, constructor: nip.constructor.Constructor):
@@ -419,19 +370,23 @@ class Nothing(Element):
     def dump(self, dumper: nip.dumper.Dumper):
         return ""
 
+    def to_python(self):
+        return None
+
 
 class FString(Element):  # Includes f-string and r-string
     @classmethod
     def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> \
             Union[FString, None]:
-        pos, string, t = tokens.PythonString.read(stream, parser.implicit_fstrings)
+        read_tokens = stream.peek(tokens.PythonString)
+        if read_tokens is None:
+            return None
+        stream.step()
+        string, t = read_tokens[0].value
         if t == 'r':
             print("Warning: all strings in NIP are already python r-string. "
                   "You don't have to explicitly specify it.")
-        if pos > 0:
-            stream.move(pos)
-            return FString(value=string)
-        return None
+        return FString(value=string)
 
     def construct(self, constructor: nip.constructor.Constructor):
         locals().update(constructor.vars)
@@ -441,4 +396,4 @@ class FString(Element):  # Includes f-string and r-string
         return f"f{self.value}"
 
     def to_python(self):
-        return f"f{self.value}"
\ No newline at end of file
+        return f"f{self.value}"
